@@ -8,8 +8,9 @@ our (@ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
 
 @ISA = qw (Exporter);
 
-@EXPORT = qw (&parse_am_file &output_am_file 
-	      &parse_configure_in parse_configure_in_buffer
+@EXPORT = qw (&debug &enable_debug
+	      &parse_am_file &output_am_file 
+	      &parse_configure_in &parse_configure_in_buffer
 	      &output_configure_in
 	      &configure_rewrite_variable &configure_remove_variable
 	      &configure_rewrite_packages
@@ -17,7 +18,7 @@ our (@ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
 	      &use_macro &use_rule &expand_one_var
 	      &macro_create &macro_remove &macro_rewrite 
 	      &macro_append_text &macro_prepend_text &macro_remove_text
-	      &rule_remove, &am_file_reset
+	      &rule_remove &am_file_reset
 	      &edit_config_files);
 @EXPORT_OK = qw (&test_print_am_file &get_line_from_hint &get_config_files);
 %EXPORT_TAGS = (DEBUG => [qw (&get_config_files &test_print_am_file)]);
@@ -29,7 +30,7 @@ use GBF::General;
 my $debug = 0;
 
 sub debug { my $message = $_[0]; print STDERR "DEBUG: $message\n" if $debug; }
-
+sub enable_debug { $debug = 1; }
 
 # +----------------
 # | makefile              : information extracted from the parsing of Makefile.am
@@ -68,38 +69,44 @@ sub debug { my $message = $_[0]; print STDERR "DEBUG: $message\n" if $debug; }
 # Constants and Variables
 ######################################################################
 
-my ($IGNORE_PATTERN, 
-    $WHITE_PATTERN,
-    $COMMENT_PATTERN,
-    $RULE_PATTERN,
-    $SUFFIX_RULE_PATTERN,
-    $MACRO_PATTERN,
-    $BOGUS_MACRO_PATTERN,
-    $IF_PATTERN,
-    $ELSE_PATTERN,
-    $ENDIF_PATTERN,
-    $PATH_PATTERN,
-    $INCLUDE_PATTERN);
-
 # String constants.
 # These are stolen from automake
-$IGNORE_PATTERN = "^##([^#].*)?\$";
-$WHITE_PATTERN = "^[ \t]*\$";
-$COMMENT_PATTERN = "^#";
-$RULE_PATTERN = "^([\$a-zA-Z_.][-+.a-zA-Z0-9_(){}/\$]*) *:([^=].*|)\$";
-$SUFFIX_RULE_PATTERN = "^\\.([a-zA-Z]+)\\.([a-zA-Z]+)\$";
-$MACRO_PATTERN = "^([A-Za-z][A-Za-z0-9_]*)[ \t]*([:+]?)=[ \t]*(.*)\$";
-$BOGUS_MACRO_PATTERN = "^([^ \t]*)[ \t]*([:+]?)=[ \t]*(.*)\$";
+
+my $IGNORE_PATTERN = '^\s*##([^#\n].*)?\n';
+my $WHITE_PATTERN = '^\s*' . "\$";
+my $COMMENT_PATTERN = '^#';
+my $TARGET_PATTERN='[$a-zA-Z_.@%][-.a-zA-Z0-9_(){}/$+@%]*';
+# A rule has three parts: a list of targets, a list of dependencies,
+# and optionally actions.
+my $RULE_PATTERN =
+  "^($TARGET_PATTERN(?:(?:\\\\\n|\\s)+$TARGET_PATTERN)*) *:([^=].*|)\$";
+
+# Only recognize leading spaces, not leading tabs.  If we recognize
+# leading tabs here then we need to make the reader smarter, because
+# otherwise it will think rules like `foo=bar; \' are errors.
+my $ASSIGNMENT_PATTERN = '^ *([^ \t=:+]*)\s*([:+]?)=\s*(.*)' . "\$";
 # This pattern recognizes a Gnits version id and sets $1 if the
 # release is an alpha release.  We also allow a suffix which can be
 # used to extend the version number with a "fork" identifier.
-$IF_PATTERN = "^if[ \t]+\([A-Za-z][A-Za-z0-9_]*\)[ \t]*\(#.*\)?\$";
-$ELSE_PATTERN = "^else[ \t]*\(#.*\)?\$";
-$ENDIF_PATTERN = "^endif[ \t]*\(#.*\)?\$";
-$PATH_PATTERN='(\\w|/|\\.)+';
-# This will pass through anything not of the prescribed form.
-$INCLUDE_PATTERN = "^include[ \t]+((\\\$\\\(top_srcdir\\\)/${PATH_PATTERN})|(\\\$\\\(srcdir\\\)/${PATH_PATTERN})|([^/\\\$]${PATH_PATTERN}))[ \t]*(#.*)?\$";
+my $GNITS_VERSION_PATTERN = '\d+\.\d+([a-z]|\.\d+)?(-[A-Za-z0-9]+)?';
 
+my $IF_PATTERN = '^if\s+(!?)\s*([A-Za-z][A-Za-z0-9_]*)\s*(?:#.*)?' . "\$";
+my $ELSE_PATTERN =
+  '^else(?:\s+(!?)\s*([A-Za-z][A-Za-z0-9_]*))?\s*(?:#.*)?' . "\$";
+my $ENDIF_PATTERN =
+  '^endif(?:\s+(!?)\s*([A-Za-z][A-Za-z0-9_]*))?\s*(?:#.*)?' . "\$";
+my $PATH_PATTERN = '(\w|[+/.-])+';
+# This will pass through anything not of the prescribed form.
+my $INCLUDE_PATTERN = ('^include\s+'
+		       . '((\$\(top_srcdir\)/' . $PATH_PATTERN . ')'
+		       . '|(\$\(srcdir\)/' . $PATH_PATTERN . ')'
+		       . '|([^/\$]' . $PATH_PATTERN . '))\s*(#.*)?' . "\$");
+
+# Match `-d' as a command-line argument in a string.
+my $DASH_D_PATTERN = "(^|\\s)-d(\\s|\$)";
+# Directories installed during 'install-exec' phase.
+my $EXEC_DIR_PATTERN =
+  '^(?:bin|sbin|libexec|sysconf|localstate|lib|pkglib|.*exec.*)' . "\$";
 
 my %default_macro_style = ( bs_column          => -1,
 			    items_per_line     => 1,
@@ -111,6 +118,24 @@ my %default_macro_style = ( bs_column          => -1,
 # Parser Functions
 ######################################################################
 
+# &check_trailing_slash ( $LINE)
+# --------------------------------------
+# Return 1 iff $LINE ends with a slash.
+# Might modify $LINE.
+sub check_trailing_slash (\$)
+{
+  my ($line) = @_;
+
+  # Ignore `##' lines.
+  return 0 if $$line =~ /$IGNORE_PATTERN/o;
+
+  # Catch and fix a common error.
+  #msg "syntax", $where, "whitespace following trailing backslash"
+  $$line =~ s/\\\s+\n$/\\\n/;
+
+  return $$line =~ /\\$/;
+}
+
 # Function parse_file 
 
 # Parses a Makefile.am file extracting the macros and rules in it, while saving
@@ -122,18 +147,22 @@ sub parse_am_file
 {
     my $filename = $_[0];
 
-    my ($continued_line, $in_rule, $is_ok_macro, $current_rule,	$current_macro);
-    my (@lines, %macros, %rules, $rule_indentation, @extra_text);
+    my (%macros, %rules, @extra_text);
 
     if (! open (MAKEFILE, "< $filename")) {
 	&report_error (100, "Can't open file $filename: $!");
 	return {};
     }
 
-    $continued_line = $in_rule = 0;
-    $current_rule = $current_macro = "";
-    @lines = ();
-    $rule_indentation = "";
+
+    use constant IN_VAR_DEF => 0;
+    use constant IN_RULE_DEF => 1;
+    use constant IN_COMMENT => 2;
+    my $prev_state = IN_RULE_DEF;
+    my $saw_bk = 0;
+    my $current_rule = "";
+    my @lines = ();
+    my $rule_indentation = "";
 
     while (<MAKEFILE>) {
 	push @lines, $_;
@@ -142,22 +171,52 @@ sub parse_am_file
 	    # Merely delete comments beginning with two hashes.
 
 	} elsif (/$WHITE_PATTERN/o) {
-	    if ($continued_line) {
+	    if ($saw_bk) {
 		&report_error (101, "$filename:$.: blank line following " .
 			       "trailing backslash");
 		return {};
 	    };
 
 	} elsif (/$COMMENT_PATTERN/o) {
-	    if ($continued_line) {
-		&report_error (101, "$filename:$.: comment following " .
+	    $prev_state = IN_COMMENT;
+	} else {
+	    last;
+	}
+	$saw_bk = check_trailing_slash ($_);
+    }
+
+    my $last_var_name = '';
+    my $last_var_type = '';
+    my $last_var_value = '';
+    # FIXME: shouldn't use $_ in this loop; it is too big.
+    while ($_) {
+	# Make sure the line is \n-terminated.
+	chomp;
+	$_ .= "\n";
+
+	my $new_saw_bk = check_trailing_slash ($_);
+
+	if (/$IGNORE_PATTERN/o) {
+	    # Merely delete comments beginning with two hashes.
+
+	    # Keep any backslash from the previous line.
+	    $new_saw_bk = $saw_bk;
+	} elsif (/$WHITE_PATTERN/o) {
+	    if ($saw_bk) {
+		&report_error (101, "$filename:$.: blank line following " .
 			       "trailing backslash");
 		return {};
 	    };
 
-	} elsif ($continued_line) {
-	    $continued_line = /\\\s*$/;
-	    if ($in_rule) {
+	} elsif (/$COMMENT_PATTERN/o) {
+	    if ($saw_bk && $prev_state != IN_COMMENT) {
+		&report_error (101, "$filename:$.: comment following " .
+			       "trailing backslash");
+		return {};
+	    };
+	    $prev_state = IN_COMMENT;
+	} elsif ($saw_bk) {	    
+	    if ($prev_state == IN_RULE_DEF) {
 		if ($rules{$current_rule}{actions} eq "") {
 		    # Strip trailing backslash from rule deps
 		    s/\\\s*$//;
@@ -165,23 +224,28 @@ sub parse_am_file
 		} else {
 		    $rules{$current_rule}{actions} .= $_;
 		};
-	    }
-	    elsif ($current_macro) {
+	    } elsif ($prev_state == IN_COMMENT) {
+	    } else # $prev_state == IN_VAR_DEF
+	    {
 		# Chop newline and backslash if this line is
 		# continued.  ensure trailing whitespace exists.
-		s/\\\s*$//;
-		$macros{$current_macro}{contents} .= $_;
+		s/\\\s*$//;		
+		$last_var_value .= ' '
+		    unless $last_var_value =~ /\s$/;
+		$last_var_value .= $_;
+		$macros{$last_var_name}{contents} = $last_var_value;
 	    }
-	    else {
-		&report_error (101, "$filename:$.: continued line but not in rule " .
-			       "or macro definition");
-		return {};
-	    };
-	    
+	} elsif (/$IF_PATTERN/o) {
+	    #$cond = cond_stack_if ($1, $2, $where);
+	} elsif (/$ELSE_PATTERN/o) {
+	    #$cond = cond_stack_else ($1, $2, $where);
+	} elsif (/$ENDIF_PATTERN/o) {
+	    #$cond = cond_stack_endif ($1, $2, $where);
 	} elsif (/$RULE_PATTERN/o) {
 	    # Found a rule.
-	    $in_rule = 1;
 	    $current_rule = $1;
+	    &debug ("$filename:$.: Found rule '$current_rule'");
+	    $prev_state = IN_RULE_DEF;
 	    $rule_indentation = "";
 
 	    my $value = $2;
@@ -190,54 +254,40 @@ sub parse_am_file
 	    } else {
 		$value = "";
 	    };
-	    $rules{$1} = { actions      => "",
-			   dependencies => $value,
-			   atline       => $.,
-			   used         => 0 };
-	    
-	    &debug ("$filename:$.: Found rule '$1'");
-	    $continued_line = /\\\s*$/;
-
-	} elsif (($is_ok_macro = /$MACRO_PATTERN/o)
-		 || /$BOGUS_MACRO_PATTERN/o) {
+	    $rules{$current_rule} = { actions      => "",
+				      dependencies => $value,
+				      atline       => $.,
+				      used         => 0 };
+	} elsif (/$ASSIGNMENT_PATTERN/o) {
 	    # Found a macro definition.
-	    my $value;
-
-	    $in_rule = 0;
-	    $current_macro = $1;
-
+	    $prev_state = IN_VAR_DEF;
+	    $last_var_name = $1;
+	    $last_var_type = $2;
 	    if ($3 ne '' && substr ($3, -1) eq "\\") {
-		$value = substr ($3, 0, length ($3) - 1);
+		$last_var_value = substr ($3, 0, length ($3) - 1);
 	    } else {
-		$value = $3;
+		$last_var_value = $3;
 	    }
 
 	    if ($2 ne '+') {
-		$macros{$1} = { contents => $value,
+		$macros{$1} = { contents => $last_var_value,
 				atline => $.,
 				used => 0};
 	    } else {
 		# Add to the previously defined var
 		if (defined ($macros{$1})) {
-		    $macros{$1}{contents} .= $value;
+		    $macros{$1}{contents} .= ' ' . $last_var_value;
 		} else {
 		    &report_warning (100, "$filename:$.: Macro '$1' not " .
 				     "previously defined");
-		    $macros{$1} = { contents => $value,
+		    $macros{$1} = { contents => $last_var_value,
 				    atline => $.,
 				    used => 0};
 		}
 	    };
-
-	    &debug ("$filename:$.: Found macro '$1'");
-	    
-	    $continued_line = /\\\s*$/;
-
 	} elsif (/$INCLUDE_PATTERN/o) {
-	    # Include
 	    &debug ("$filename:$.: Found include, FIXME: handle it!");
-
-	} else {
+        } else {
 	    # This isn't an error; it is probably a continued rule.
 	    # In fact, this is what we assume.
 
@@ -249,19 +299,25 @@ sub parse_am_file
 		    return {};
 		};
 	    } else {
-		$in_rule = 0;
+		$prev_state = -1;
 	    }
 
-	    $current_macro = "";
-	    $continued_line = /\\\s*$/;
+	    $last_var_name = "";
 	    
-	    if ($in_rule) {
+	    if ($prev_state == IN_RULE_DEF) {
 		$rules{$current_rule}{actions} .= $_;
 	    } else {
 		push @extra_text, $_;
 	    };
 	}
+
+	$saw_bk = $new_saw_bk;
+        $_ = <MAKEFILE>;
+	push @lines, $_;
     };
+
+    &report_error (100, "trailing backslash on last line")
+	if $saw_bk;    
     
     close MAKEFILE;
 
@@ -477,7 +533,7 @@ sub configure_rewrite_packages
     } else {
 	## Could not find any proper place to put the new variable.
 	## Just putting it on top.
-	$configure_in->{'contents'} =~ s/AC_OUTPUT/PKG_CHECK_MODULES($module, $packages)\n\nAC_OUTPUT/sm;
+	$configure_in->{'contents'} =~ s/AC_OUTPUT/PKG_CHECK_MODULES($module, $packages)\nAC_SUBST(${module}_CFLAGS)\nAC_SUBST(${module}_LIBS)\n\nAC_OUTPUT/sm;
     }
     $configure_in->{"dirty"} = 1;
 }
@@ -744,7 +800,7 @@ sub analyze_macro_style
 	};
 	
 	if ($first_line) {
-	    $line =~ /$MACRO_PATTERN/;
+	    $line =~ /$ASSIGNMENT_PATTERN/;
 	    $line_copy = &empty_if_undef ($3);
 	} elsif (!$leading_space) {
 	    $line =~ /^(\s+)\S+/;
@@ -807,7 +863,7 @@ sub format_macro
 	$leading_space = "\t";
     }
 
-    if ($line =~ /$MACRO_PATTERN/) {
+    if ($line =~ /$ASSIGNMENT_PATTERN/) {
 	my @current_items = split /\s+/, &trim (&empty_if_undef ($3));
 	$item_count = scalar @current_items;
 
@@ -923,7 +979,7 @@ sub macro_prepend_text
 	my $first_line_num = $macro->{atline};
 	my $first_line = @{$makefile->{lines}}[$first_line_num - 1];
 
-	$first_line =~ /$MACRO_PATTERN/;
+	$first_line =~ /$ASSIGNMENT_PATTERN/;
 	my $leading_text = $3;
 	$text .= " " . $leading_text;
 	$text =~ s/\\\s*$//;
