@@ -29,6 +29,7 @@
 #include <glade/glade-xml.h>
 #include <gtk/gtk.h>
 #include <libgnomevfs/gnome-vfs-uri.h>
+#include <libgnomevfs/gnome-vfs-utils.h>
 #include <bonobo/bonobo-file-selector-util.h>
 #include <gdl/gdl-icons.h>
 
@@ -42,6 +43,9 @@
 
 #define GLADE_FILE GBF_GLADEDIR "/create_dialogs.glade"
 
+#define COLUMN_FILE 0
+#define COLUMN_URI 1
+
 static GtkWidget *
 custom_widget_handler (GladeXML *xml,
                        gchar    *func_name,
@@ -142,25 +146,6 @@ setup_groups_treeview (GbfProjectModel *model,
     gtk_tree_path_free (path);
 }
 
-static void
-entry_changed_cb (GtkEditable *editable, gpointer user_data)
-{
-    GtkWidget *button = user_data;
-    gchar *text;
-
-    if (!button)
-        return;
-    
-    text = gtk_editable_get_chars (editable, 0, -1);
-    if (strlen (text) > 0) {
-        gtk_widget_set_sensitive (button, TRUE);
-        gtk_widget_grab_default (button);
-    } else {
-        gtk_widget_set_sensitive (button, FALSE);
-    }
-    g_free (text);
-}
-
 static void
 error_dialog (GtkWindow *parent, const gchar *msg, ...)
 {
@@ -183,6 +168,26 @@ error_dialog (GtkWindow *parent, const gchar *msg, ...)
     gtk_widget_destroy (dialog);
 }
 
+static void
+entry_changed_cb (GtkEditable *editable, gpointer user_data)
+{
+    GtkWidget *button = user_data;
+    gchar *text;
+
+    if (!button)
+        return;
+
+    text = gtk_editable_get_chars (editable, 0, -1);
+    if (strlen (text) > 0) {
+        gtk_widget_set_sensitive (button, TRUE);
+        gtk_widget_grab_default (button);
+    } else {
+        gtk_widget_set_sensitive (button, FALSE);
+    }
+    g_free (text);
+}
+
+
 gchar*
 gbf_project_util_new_group (GbfProjectModel *model,
                             GtkWindow       *parent,
@@ -534,16 +539,52 @@ setup_targets_treeview (GbfProjectModel *model,
     }
 }
 
+static void
+clean_model(GtkTreeModel* model)
+{
+	GtkTreeIter iter;
+	if (!gtk_tree_model_get_iter_first(model, &iter))
+		return;
+	do
+	{
+		gchar* source_file;
+    	gchar* filename;
+    	gtk_tree_model_get (model, &iter, COLUMN_FILE, &filename,
+        	 COLUMN_URI, &source_file, -1);        
+		g_free(filename);
+		g_free(source_file);
+	} while (gtk_tree_model_iter_next(GTK_TREE_MODEL(model), &iter));
+	gtk_list_store_clear(GTK_LIST_STORE(model));
+}
+
 static void
 browse_button_clicked_cb (GtkWidget *widget, gpointer user_data)
 {
-    GtkEntry *entry = user_data;
-    gchar *uri, *selected_uri;
+    GtkTreeView *tree = user_data;
+    gchar *file, *uri;
     GnomeVFSURI *tmp_uri;
+    GtkFileChooserDialog* dialog;
+	GtkTreeModel* model;
+    GtkTreeIter iter;
+    gint result;
     
-    g_return_if_fail (user_data != NULL && GTK_IS_ENTRY (user_data));
-
-    uri = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
+    g_return_if_fail (user_data != NULL && GTK_IS_TREE_VIEW (user_data));
+	
+	model = gtk_tree_view_get_model(tree);
+	if (gtk_tree_model_get_iter_first(model, &iter))
+	{
+		gtk_tree_model_get(model, &iter, COLUMN_URI, &file, -1);
+		uri = g_strdup(file);
+	}
+	else
+		uri = g_strdup("");
+	
+    dialog = GTK_FILE_CHOOSER_DIALOG(gtk_file_chooser_dialog_new (_("Select sources..."),
+				      GTK_WINDOW (gtk_widget_get_toplevel (widget)),
+				      GTK_FILE_CHOOSER_ACTION_OPEN,
+				      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+				      GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
+				      NULL));
     tmp_uri = gnome_vfs_uri_new (uri);
     g_free (uri);
     uri = NULL;
@@ -551,34 +592,100 @@ browse_button_clicked_cb (GtkWidget *widget, gpointer user_data)
         uri = gnome_vfs_uri_extract_dirname (tmp_uri);
         gnome_vfs_uri_unref (tmp_uri);
     }
-    selected_uri = bonobo_file_selector_open (
-        GTK_WINDOW (gtk_widget_get_toplevel (widget)),
-        FALSE, _("Add source..."), NULL,
-        uri ? uri : g_object_get_data (G_OBJECT (widget), "root"));
+    
+    gtk_file_chooser_set_current_folder_uri(GTK_FILE_CHOOSER(dialog),
+    	uri ? uri : g_object_get_data (G_OBJECT (widget), "root"));
+    gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(dialog), TRUE);
     g_free (uri);
 
-    if (selected_uri) {
-        gtk_entry_set_text (entry, selected_uri);
-        g_free (selected_uri);
-    }
+	result = gtk_dialog_run(GTK_DIALOG(dialog));
+	switch(result)
+	{
+		case GTK_RESPONSE_ACCEPT:
+		{
+			GSList* uris = gtk_file_chooser_get_uris(GTK_FILE_CHOOSER(dialog));
+			GSList* filenames = gtk_file_chooser_get_filenames(GTK_FILE_CHOOSER(dialog));
+			GSList* node_filename = filenames;
+			GSList* node_uri = uris;
+			clean_model(model);
+			while (node_uri != NULL && node_filename != NULL)
+			{
+				GtkTreeIter iter;
+				gchar* uri = node_uri->data;
+				gchar* file = node_filename->data;
+				gtk_list_store_append(GTK_LIST_STORE(model), &iter);
+				gtk_list_store_set(GTK_LIST_STORE(model), &iter, COLUMN_FILE, file, COLUMN_URI, uri, -1);
+				node_uri = g_slist_next(node_uri);
+				node_filename = g_slist_next(node_filename);
+			}
+  			g_slist_free(uris);
+  			g_slist_free(filenames);
+  			break;
+  		}
+  		default:
+  			break;
+  	} 
+  	gtk_widget_destroy (GTK_WIDGET(dialog));
 }
 
-gchar* 
+static void
+on_row_changed(GtkTreeModel* model, GtkTreePath* path, GtkTreeIter* iter, gpointer data)
+{
+	GtkWidget* button = GTK_WIDGET(data);
+	if (gtk_list_store_iter_is_valid(GTK_LIST_STORE(model), iter))
+		gtk_widget_set_sensitive(button, TRUE);
+	else
+		gtk_widget_set_sensitive(button, FALSE);
+}
+
+gchar*
 gbf_project_util_add_source (GbfProjectModel *model,
                              GtkWindow       *parent,
                              const gchar     *default_target,
                              const gchar     *default_group,
                              const gchar     *default_uri)
+{
+	gchar* uri = g_strdup(default_uri);
+	GList* uris = g_list_append(NULL, uri);
+	GList* new_sources = 
+		gbf_project_util_add_source_multi(model, parent, default_target, default_group, uris);
+	g_free(uri);
+	
+	if (g_list_length(new_sources))
+	{
+		gchar* new_source = g_strdup(new_sources->data);
+		while(new_sources) 
+		{
+			g_free(new_sources->data);
+			new_sources = g_list_next(new_sources);
+		}
+		g_list_free(new_sources);
+		return new_source;
+	}
+	else
+		return NULL;
+}
+
+GList* 
+gbf_project_util_add_source_multi (GbfProjectModel *model,
+				    GtkWindow       *parent,
+				    const gchar     *default_target,
+				    const gchar     *default_group,
+				    GList     *uris_to_add)
 {
     GladeXML *gui;
-    GtkWidget *dialog, *source_file_entry;
+    GtkWidget *dialog, *source_file_tree;
     GtkWidget *ok_button, *browse_button;
     GtkWidget *targets_view;
     gint response;
     GbfProject *project;
     gboolean finished = FALSE;
     gchar *project_root;
-    gchar *new_source = NULL;
+    GtkListStore* list;
+    GtkCellRenderer* renderer;
+    GtkTreeViewColumn* column_filename;
+    GList* new_sources = NULL;
+    GList* uri_node;
     
     g_return_val_if_fail (model != NULL, NULL);
     
@@ -592,21 +699,41 @@ gbf_project_util_add_source (GbfProjectModel *model,
     /* get all needed widgets */
     dialog = glade_xml_get_widget (gui, "add_source_dialog");
     targets_view = glade_xml_get_widget (gui, "targets_view");
-    source_file_entry = glade_xml_get_widget (gui, "source_file_entry");
+    source_file_tree = glade_xml_get_widget (gui, "source_file_tree");
     browse_button = glade_xml_get_widget (gui, "browse_button");
     ok_button = glade_xml_get_widget (gui, "ok_button");
     
+    /* Prepare file tree */
+    list = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
+    gtk_tree_view_set_model(GTK_TREE_VIEW(source_file_tree), GTK_TREE_MODEL(list));
+    renderer = gtk_cell_renderer_text_new();
+    column_filename = gtk_tree_view_column_new_with_attributes("Files", renderer, "text", COLUMN_FILE, NULL);
+    gtk_tree_view_column_set_sizing(column_filename, GTK_TREE_VIEW_COLUMN_FIXED);
+    gtk_tree_view_append_column(GTK_TREE_VIEW(source_file_tree), column_filename);
+    
     /* set up dialog */
-    g_signal_connect (source_file_entry, "changed",
-                      (GCallback) entry_changed_cb, ok_button);
-    if (default_uri) {
-        gtk_entry_set_text (GTK_ENTRY (source_file_entry), default_uri);
-        gtk_widget_set_sensitive (ok_button, TRUE);
-    } else {
-        gtk_widget_set_sensitive (ok_button, FALSE);
+   uri_node = uris_to_add;
+   while (uri_node) 
+   {
+        GtkTreeIter iter;
+        gchar* filename = gnome_vfs_get_local_path_from_uri(uri_node->data);
+        if (!filename)
+        	filename = g_strdup(uri_node->data);
+        gtk_list_store_append(list, &iter);
+        gtk_list_store_set(list, &iter, COLUMN_FILE, filename,
+        	COLUMN_URI, g_strdup(uri_node->data), -1);
+    
+    	uri_node = g_list_next(uri_node);
     }
+    if (!g_list_length(uris_to_add))
+   		gtk_widget_set_sensitive (ok_button, FALSE);
+   	else
+   		gtk_widget_set_sensitive (ok_button, TRUE);
+
+    g_signal_connect(G_OBJECT(list), "row_changed", G_CALLBACK(on_row_changed), ok_button);
+    
     g_signal_connect (browse_button, "clicked",
-                      (GCallback) browse_button_clicked_cb, source_file_entry);
+                      (GCallback) browse_button_clicked_cb, source_file_tree);
     g_object_get (project, "project-dir", &project_root, NULL);
     g_object_set_data_full (G_OBJECT (browse_button), "root",
                             project_root, g_free);
@@ -619,7 +746,7 @@ gbf_project_util_add_source (GbfProjectModel *model,
     }
 
     if (default_target)
-        gtk_widget_grab_focus (source_file_entry);
+        gtk_widget_grab_focus (source_file_tree);
     else
         gtk_widget_grab_focus (targets_view);
     
@@ -630,28 +757,43 @@ gbf_project_util_add_source (GbfProjectModel *model,
         switch (response) {
             case GTK_RESPONSE_OK: 
             {
-                GError *err = NULL;
                 GbfTreeData *data;
-                gchar *target_id = NULL, *source_file;
+                gchar *target_id = NULL;
             
                 data = gbf_project_view_find_selected (GBF_PROJECT_VIEW (targets_view),
                                                        GBF_TREE_NODE_TARGET);
                 if (data) {
+                	GtkTreeIter iter;
                     gtk_widget_hide (dialog);
                     
                     target_id = g_strdup (data->id);
                     gbf_tree_data_free (data);
-                    source_file = gtk_editable_get_chars (
-                        GTK_EDITABLE (source_file_entry), 0, -1);
-
-                    new_source = gbf_project_add_source (project, target_id,
+                   
+                   	if (!gtk_tree_model_get_iter_first(GTK_TREE_MODEL(list), &iter))
+                   		break;
+                   	do
+                   	{
+                   		GError *err = NULL;
+						gchar* new_source;
+                   		gchar* source_file;
+                   		gchar* filename;
+                   		gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COLUMN_FILE, &filename,
+                   			COLUMN_URI, &source_file, -1);
+                   		
+                    	new_source = gbf_project_add_source (project, target_id,
                                                          source_file, &err);
-                    if (err) {
-                        error_dialog (parent, _("Can't add source: %s"),
-                                      err->message);
-                        g_error_free (err);
-                    }
-                    g_free (source_file);
+                    	if (err) {
+                        	error_dialog (parent, _("Can't add source: %s"),
+                            	          err->message);
+                        	g_error_free (err);
+                    	}
+						else
+							new_sources = g_list_append(new_sources, new_source);
+						
+						g_free(filename);
+                        g_free(source_file);
+                    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(list), &iter));
+                    
                     g_free (target_id);
                     finished = TRUE;
                     
@@ -663,6 +805,7 @@ gbf_project_util_add_source (GbfProjectModel *model,
                 break;
             }
             default:
+            	clean_model(GTK_TREE_MODEL(list));
                 finished = TRUE;
                 break;
         }
@@ -671,5 +814,5 @@ gbf_project_util_add_source (GbfProjectModel *model,
     /* destroy stuff */
     gtk_widget_destroy (dialog);
     g_object_unref (gui);
-    return new_source;
+    return new_sources;
 }
