@@ -171,6 +171,8 @@ struct _GbfProjectClass {
 	GbfProjectGroup *        (* get_group)             (GbfProject  *project,
 							    const gchar *id,
 							    GError     **error);
+	GList *                  (* get_all_groups)        (GbfProject  *project,
+							    GError     **error);
 	GtkWidget *              (* configure_group)       (GbfProject  *project,
 							    const gchar *id,
 							    GError     **error);
@@ -215,6 +217,8 @@ struct _GbfProjectClass {
 							    GError     **error);
 	GtkWidget *              (* configure_new_source)  (GbfProject  *project,
 							    GError     **error);
+	GtkWidget *              (* configure)             (GbfProject  *project,
+							    GError     **error);
 
 	/* Types. */
 	const gchar *            (* name_for_type)         (GbfProject  *project,
@@ -267,6 +271,8 @@ void                    gbf_project_remove_group         (GbfProject    *project
 GbfProjectGroup        *gbf_project_get_group            (GbfProject    *project,
 							  const gchar   *id,
 							  GError       **error);
+GList                  *gbf_project_get_all_groups       (GbfProject    *project,
+							  GError       **error);
 GtkWidget              *gbf_project_configure_group      (GbfProject    *project,
 							  const gchar   *id,
 							  GError       **error);
@@ -313,7 +319,10 @@ GtkWidget              *gbf_project_configure_source     (GbfProject    *project
 							  GError       **error);
 GtkWidget              *gbf_project_configure_new_source (GbfProject    *project,
 							  GError       **error);
+/* Project */
 
+GtkWidget              *gbf_project_configure            (GbfProject *project,
+							  GError **error);
 
 /* Types. */
 const gchar            *gbf_project_name_for_type        (GbfProject    *project,
